#include <open62541/client_subscriptions.h>
#include <open62541/server_config_default.h>
#include <open62541/types_generated.h>
#include <open62541/plugin/pubsub_udp.h>

#include "ua_pubsub.h"
#include "pubsub.h"

UA_NodeId connectionIdent, publishedDataSetIdent, writerGroupIdent,
          readerGroupIdent, readerIdent;

UA_DataSetReaderConfig readerConfig;

const UA_Logger *pubsubLogger;

VariableData (*pubsubGetValue)(UA_String identifier);
static void (*callbackUpdate)(UA_UInt32, const UA_DataValue*, bool print);

static void fillDataSetMetaData(UA_DataSetMetaDataType *pMetaData,
                                VariableStruct variables, int id);

static UA_StatusCode
addPubSubConnection(UA_Server *server, UA_String *transportProfile,
                    UA_NetworkAddressUrlDataType *networkAddressUrl,
                    UA_UInt32 id){
    // 18 characters for "UDPMC Connection " + 10 characters for id <= UINT32_MAX
    char name[28];
    if((server == NULL) || (transportProfile == NULL) ||
        (networkAddressUrl == NULL)) {
        return UA_STATUSCODE_BADINTERNALERROR;
    }

    UA_PubSubConnectionConfig connectionConfig;
    memset(&connectionConfig, 0, sizeof(connectionConfig));

    UA_snprintf(name, sizeof(name) - 1, "UDPMC Connection %d", id);
    connectionConfig.name = UA_STRING(name);

    connectionConfig.transportProfileUri = *transportProfile;
    connectionConfig.enabled = UA_TRUE;
    UA_Variant_setScalar(&connectionConfig.address, networkAddressUrl,
                         &UA_TYPES[UA_TYPES_NETWORKADDRESSURLDATATYPE]);
    connectionConfig.publisherId.numeric = id;
    return UA_Server_addPubSubConnection(server, &connectionConfig, &connectionIdent);
}

/**
 * **PublishedDataSet handling**
 *
 * The PublishedDataSet (PDS) and PubSubConnection are the toplevel entities and
 * can exist alone. The PDS contains the collection of the published fields. All
 * other PubSub elements are directly or indirectly linked with the PDS or
 * connection. */
static void
addPublishedDataSet(UA_Server *server, UA_UInt32 id) {
    /* The PublishedDataSetConfig contains all necessary public
    * information for the creation of a new PublishedDataSet */
    char name[8];
    UA_snprintf(name, sizeof(name) - 1, "PDS %d", id);
    UA_PublishedDataSetConfig publishedDataSetConfig;
    memset(&publishedDataSetConfig, 0, sizeof(UA_PublishedDataSetConfig));
    publishedDataSetConfig.publishedDataSetType = UA_PUBSUB_DATASET_PUBLISHEDITEMS;
    publishedDataSetConfig.name = UA_STRING(name);
    /* Create new PublishedDataSet based on the PublishedDataSetConfig. */
    UA_Server_addPublishedDataSet(server, &publishedDataSetConfig, &publishedDataSetIdent);
}

static UA_StatusCode
readVariable(UA_Server *server,
             const UA_NodeId *sessionId, void *sessionContext,
             const UA_NodeId *nodeId, void *nodeContext,
             UA_Boolean sourceTimeStamp, const UA_NumericRange *range,
             UA_DataValue *dataValue) {
    UA_StatusCode retval;
    VariableData varDetails = pubsubGetValue(nodeId->identifier.string);

    switch(varDetails.valueRank) {
        case UA_VALUERANK_SCALAR:
            retval = UA_Variant_setScalarCopy(&dataValue->value,
                                              varDetails.value,
                                              &UA_TYPES[varDetails.type]);
            break;
        case UA_VALUERANK_ONE_DIMENSION:
            retval = UA_Variant_setArrayCopy(&dataValue->value,
                                             varDetails.value,
                                             varDetails.arrayDimensions[0],
                                             &UA_TYPES[varDetails.type]);
            break;
        default:
            retval = UA_STATUSCODE_BAD;
            break;
    }

    if(retval != UA_STATUSCODE_GOOD)
        UA_LOG_ERROR(pubsubLogger, UA_LOGCATEGORY_USERLAND, "read returned value %d", retval);
    dataValue->hasValue = true;
    return retval;
}

static UA_StatusCode
writeVariable(UA_Server *server,
              const UA_NodeId *sessionId, void *sessionContext,
              const UA_NodeId *nodeId, void *nodeContext,
              const UA_NumericRange *range, const UA_DataValue *data) {
    UA_LOG_INFO(pubsubLogger, UA_LOGCATEGORY_USERLAND,
                "Updating variables manually is not implemented");
    return UA_STATUSCODE_BADINTERNALERROR;
}

static UA_StatusCode
addDataSourceVariable(UA_Server *server, VariableData varDetails) {
    UA_VariableAttributes attr = UA_VariableAttributes_default;
    attr.description = UA_LOCALIZEDTEXT("en-US", varDetails.description);
    attr.displayName = UA_LOCALIZEDTEXT("en-US", varDetails.name);
    attr.dataType = UA_TYPES[varDetails.type].typeId;
    attr.valueRank = varDetails.valueRank;
    attr.arrayDimensions = varDetails.arrayDimensions;
    attr.arrayDimensionsSize = varDetails.arrayDimensionsSize;
    attr.accessLevel = UA_ACCESSLEVELMASK_READ | UA_ACCESSLEVELMASK_WRITE;

    UA_DataSource dataSource;
    dataSource.read = readVariable;
    dataSource.write = writeVariable;
    return UA_Server_addDataSourceVariableNode(server,
                                               UA_NODEID_STRING(1, varDetails.name),
                                               UA_NODEID_NUMERIC(0, UA_NS0ID_OBJECTSFOLDER),
                                               UA_NODEID_NUMERIC(0, UA_NS0ID_ORGANIZES),
                                               UA_QUALIFIEDNAME(1, varDetails.name),
                                               varDetails.typeNodeId,
                                               attr, dataSource, NULL, NULL);
}

static void
addDataSetField(UA_Server *server, VariableData varDetails) {
    UA_NodeId dataSetFieldIdent;
    UA_DataSetFieldConfig dataSetFieldConfig;
    memset(&dataSetFieldConfig, 0, sizeof(UA_DataSetFieldConfig));
    dataSetFieldConfig.dataSetFieldType = UA_PUBSUB_DATASETFIELD_VARIABLE;
    dataSetFieldConfig.field.variable.fieldNameAlias = UA_STRING(varDetails.name);
    dataSetFieldConfig.field.variable.promotedField = UA_FALSE;
    dataSetFieldConfig.field.variable.publishParameters.publishedVariable =
    UA_NODEID_STRING(1, varDetails.name);
    dataSetFieldConfig.field.variable.publishParameters.attributeId = UA_ATTRIBUTEID_VALUE;
    UA_Server_addDataSetField(server, publishedDataSetIdent,
                              &dataSetFieldConfig, &dataSetFieldIdent);
}

/**
 * **WriterGroup handling**
 *
 * The WriterGroup (WG) is part of the connection and contains the primary
 * configuration parameters for the message creation. */
static void
addWriterGroup(UA_Server *server, UA_Duration publishingInterval) {
    /* Now we create a new WriterGroupConfig and add the group to the existing
     * PubSubConnection. */
    UA_WriterGroupConfig writerGroupConfig;
    memset(&writerGroupConfig, 0, sizeof(UA_WriterGroupConfig));
    writerGroupConfig.name = UA_STRING("Demo WriterGroup");
    writerGroupConfig.publishingInterval = publishingInterval;
    writerGroupConfig.enabled = UA_FALSE;
    writerGroupConfig.writerGroupId = WRITER_GROUP_ID;
    writerGroupConfig.encodingMimeType = UA_PUBSUB_ENCODING_UADP;
    writerGroupConfig.messageSettings.encoding             = UA_EXTENSIONOBJECT_DECODED;
    writerGroupConfig.messageSettings.content.decoded.type = &UA_TYPES[UA_TYPES_UADPWRITERGROUPMESSAGEDATATYPE];
    /* The configuration flags for the messages are encapsulated inside the
     * message- and transport settings extension objects. These extension
     * objects are defined by the standard. e.g.
     * UadpWriterGroupMessageDataType */
    UA_UadpWriterGroupMessageDataType *writerGroupMessage  = UA_UadpWriterGroupMessageDataType_new();
    /* Change message settings of writerGroup to send PublisherId,
     * WriterGroupId in GroupHeader and DataSetWriterId in PayloadHeader
     * of NetworkMessage */
    writerGroupMessage->networkMessageContentMask          = (UA_UadpNetworkMessageContentMask)(UA_UADPNETWORKMESSAGECONTENTMASK_PUBLISHERID |
                                                              (UA_UadpNetworkMessageContentMask)UA_UADPNETWORKMESSAGECONTENTMASK_GROUPHEADER |
                                                              (UA_UadpNetworkMessageContentMask)UA_UADPNETWORKMESSAGECONTENTMASK_WRITERGROUPID |
                                                              (UA_UadpNetworkMessageContentMask)UA_UADPNETWORKMESSAGECONTENTMASK_PAYLOADHEADER);
    writerGroupConfig.messageSettings.content.decoded.data = writerGroupMessage;
    UA_Server_addWriterGroup(server, connectionIdent, &writerGroupConfig, &writerGroupIdent);
    UA_Server_setWriterGroupOperational(server, writerGroupIdent);
    UA_UadpWriterGroupMessageDataType_delete(writerGroupMessage);
}

/**
 * **DataSetWriter handling**
 *
 * A DataSetWriter (DSW) is the glue between the WG and the PDS. The DSW is
 * linked to exactly one PDS and contains additional information for the
 * message generation. */
static UA_StatusCode
addDataSetWriter(UA_Server *server) {
    /* We need now a DataSetWriter within the WriterGroup. This means we must
     * create a new DataSetWriterConfig and add call the addWriterGroup function. */
    UA_NodeId dataSetWriterIdent;
    UA_DataSetWriterConfig dataSetWriterConfig;
    memset(&dataSetWriterConfig, 0, sizeof(UA_DataSetWriterConfig));
    dataSetWriterConfig.name = UA_STRING("Demo DataSetWriter");
    dataSetWriterConfig.dataSetWriterId = DATA_SET_WRITER_ID;
    dataSetWriterConfig.keyFrameCount = 10;
    return UA_Server_addDataSetWriter(server, writerGroupIdent, publishedDataSetIdent,
                                      &dataSetWriterConfig, &dataSetWriterIdent);
}

/**
 * **ReaderGroup**
 *
 * ReaderGroup is used to group a list of DataSetReaders. All ReaderGroups are
 * created within a PubSubConnection and automatically deleted if the connection
 * is removed. All network message related filters are only available in the DataSetReader. */
/* Add ReaderGroup to the created connection */
static UA_StatusCode
addReaderGroup(UA_Server *server) {
    if(server == NULL) {
        return UA_STATUSCODE_BADINTERNALERROR;
    }

    UA_StatusCode retval;
    UA_ReaderGroupConfig readerGroupConfig;
    memset (&readerGroupConfig, 0, sizeof(UA_ReaderGroupConfig));
    readerGroupConfig.name = UA_STRING("Demo ReaderGroup");
    retval = UA_Server_addReaderGroup(server, connectionIdent, &readerGroupConfig,
                                       &readerGroupIdent);
    UA_Server_setReaderGroupOperational(server, readerGroupIdent);
    return retval;
}

/**
 * **DataSetReader**
 *
 * DataSetReader can receive NetworkMessages with the DataSetMessage
 * of interest sent by the Publisher. DataSetReader provides
 * the configuration necessary to receive and process DataSetMessages
 * on the Subscriber side. DataSetReader must be linked with a
 * SubscribedDataSet and be contained within a ReaderGroup. */
/* Add DataSetReader to the ReaderGroup */
static UA_StatusCode
addDataSetReader(UA_Server *server, VariableStruct variables, int id) {
    if(server == NULL)
        return UA_STATUSCODE_BADINTERNALERROR;

    /* Setting up Meta data configuration in DataSetReader */
    fillDataSetMetaData(&readerConfig.dataSetMetaData, variables, id);

    return UA_Server_addDataSetReader(server, readerGroupIdent,
                                      &readerConfig, &readerIdent);
}

static void
dataChangeNotificationCallback(UA_Server *server, UA_UInt32 monitoredItemId,
                               void *monitoredItemContext, const UA_NodeId *nodeId,
                               void *nodeContext, UA_UInt32 attributeId,
                               const UA_DataValue *var) {
    callbackUpdate(nodeId->identifier.numeric, var, false);
}

/**
 * **SubscribedDataSet**
 *
 * Set SubscribedDataSet type to TargetVariables data type.
 * Add subscribedvariables to the DataSetReader */
static UA_StatusCode
addSubscribedVariables(UA_Server *server, UA_NodeId dataSetReaderId,
                       VariableData *variableArray, UA_UInt32 nb,
                       UA_UInt32 maxVariableNb, UA_Duration samplingInterval,
                       void (*init_node_id)(UA_UInt32 id, UA_UInt32 nb, UA_UInt32 magic)) {
    if(server == NULL)
        return UA_STATUSCODE_BADINTERNALERROR;

    UA_StatusCode retval = UA_STATUSCODE_GOOD;
    UA_NodeId folderId;
    UA_String folderName = readerConfig.dataSetMetaData.name;
    UA_ObjectAttributes oAttr = UA_ObjectAttributes_default;
    UA_QualifiedName folderBrowseName;
    if(folderName.length > 0) {
        oAttr.displayName.locale = UA_STRING ("en-US");
        oAttr.displayName.text = folderName;
        folderBrowseName.namespaceIndex = 1;
        folderBrowseName.name = folderName;
    }
    else {
        oAttr.displayName = UA_LOCALIZEDTEXT ("en-US", "Subscribed Variables");
        folderBrowseName = UA_QUALIFIEDNAME (1, "Subscribed Variables");
    }

    UA_Server_addObjectNode (server, UA_NODEID_NULL,
                             UA_NODEID_NUMERIC (0, UA_NS0ID_OBJECTSFOLDER),
                             UA_NODEID_NUMERIC (0, UA_NS0ID_ORGANIZES),
                             folderBrowseName, UA_NODEID_NUMERIC (0,
                             UA_NS0ID_BASEOBJECTTYPE), oAttr, NULL, &folderId);

/**
 * **TargetVariables**
 *
 * The SubscribedDataSet option TargetVariables defines a list of Variable mappings between
 * received DataSet fields and target Variables in the Subscriber AddressSpace.
 * The values subscribed from the Publisher are updated in the value field of these variables */
    /* Create the TargetVariables with respect to DataSetMetaData fields */
    UA_FieldTargetVariable *targetVars = (UA_FieldTargetVariable *)
            UA_calloc(readerConfig.dataSetMetaData.fieldsSize, sizeof(UA_FieldTargetVariable));
    for(UA_UInt32 i = 0; i < readerConfig.dataSetMetaData.fieldsSize; i++) {
        /* Variable to subscribe data */
        UA_VariableAttributes vAttr = UA_VariableAttributes_default;
        UA_LocalizedText_copy(&readerConfig.dataSetMetaData.fields[i].description,
                              &vAttr.description);
        vAttr.displayName = UA_LOCALIZEDTEXT("en-US", variableArray[i].name);
        vAttr.dataType = readerConfig.dataSetMetaData.fields[i].dataType;
        vAttr.valueRank = readerConfig.dataSetMetaData.fields[i].valueRank;
        vAttr.arrayDimensions = readerConfig.dataSetMetaData.fields[i].arrayDimensions;
        vAttr.arrayDimensionsSize = readerConfig.dataSetMetaData.fields[i].arrayDimensionsSize;

        UA_NodeId newNode;
        retval |= UA_Server_addVariableNode(server,
                                            UA_NODEID_NUMERIC(1, (UA_UInt32)maxVariableNb*nb + i + 50000),
                                            folderId,
                                            UA_NODEID_NUMERIC(0, UA_NS0ID_HASCOMPONENT),
                                            UA_QUALIFIEDNAME(1, (char *)readerConfig.dataSetMetaData.fields[i].name.data),
                                            variableArray[i].typeNodeId,
                                            vAttr, NULL, &newNode);

        /*monitor variable*/
        UA_MonitoredItemCreateRequest monRequest = UA_MonitoredItemCreateRequest_default(newNode);
        init_node_id(newNode.identifier.numeric, nb, i);
        monRequest.requestedParameters.samplingInterval = samplingInterval;
        UA_Server_createDataChangeMonitoredItem(server, UA_TIMESTAMPSTORETURN_SOURCE,
                                                monRequest, NULL, dataChangeNotificationCallback);

        /* For creating Targetvariables */
        UA_FieldTargetDataType_init(&targetVars[i].targetVariable);
        targetVars[i].targetVariable.attributeId  = UA_ATTRIBUTEID_VALUE;
        targetVars[i].targetVariable.targetNodeId = newNode;
    }

    retval |= UA_Server_DataSetReader_createTargetVariables(server, dataSetReaderId,
                                                            readerConfig.dataSetMetaData.fieldsSize, targetVars);
    for(size_t i = 0; i < readerConfig.dataSetMetaData.fieldsSize; i++)
        UA_FieldTargetDataType_clear(&targetVars[i].targetVariable);

    UA_free(targetVars);
    UA_free(readerConfig.dataSetMetaData.fields);
    return retval;
}

/**
 * **DataSetMetaData**
 *
 * The DataSetMetaData describes the content of a DataSet. It provides the information necessary to decode
 * DataSetMessages on the Subscriber side. DataSetMessages received from the Publisher are decoded into
 * DataSet and each field is updated in the Subscriber based on datatype match of TargetVariable fields of Subscriber
 * and PublishedDataSetFields of Publisher */
/* Define MetaData for TargetVariables */
static void fillDataSetMetaData(UA_DataSetMetaDataType *pMetaData,
                                VariableStruct variables, int id) {
    char name[12];
    UA_snprintf(name, sizeof(name) - 1, "DataSet %d", id);

    if(pMetaData == NULL)
        return;

    UA_DataSetMetaDataType_init (pMetaData);
    pMetaData->name = UA_STRING (name);

    /* Definition of number of fields sizeto create different
     * targetVariables of distinct datatype */
    pMetaData->fieldsSize = variables.nbVariable;
    UA_LOG_INFO(pubsubLogger, UA_LOGCATEGORY_USERLAND, "fieldsSize %d", (int) pMetaData->fieldsSize);
    pMetaData->fields = (UA_FieldMetaData*)UA_Array_new (pMetaData->fieldsSize,
                         &UA_TYPES[UA_TYPES_FIELDMETADATA]);

    for(size_t i = 0; i < pMetaData->fieldsSize; i++) {
        UA_FieldMetaData_init (&pMetaData->fields[i]);
 
        UA_NodeId_copy(&UA_TYPES[variables.variableArray[i].type].typeId, &pMetaData->fields[i].dataType);
        pMetaData->fields[i].builtInType = variables.variableArray[i].builtInType;
        pMetaData->fields[i].name = UA_STRING (variables.variableArray[i].name);
        pMetaData->fields[i].valueRank = variables.variableArray[i].valueRank;
        pMetaData->fields[i].arrayDimensions = variables.variableArray[i].arrayDimensions;
        pMetaData->fields[i].arrayDimensionsSize = variables.variableArray[i].arrayDimensionsSize;
    }
}

static void
setVariableType(UA_Server *server, VariableStruct variables) {
    VariableData vDetails;
    UA_VariableTypeAttributes vtAttr;

    for(UA_UInt32 i = 0; i < variables.nbVariable; i++) {
        vDetails = variables.variableArray[i];

        switch(vDetails.valueRank) {
            case UA_VALUERANK_SCALAR:
                vDetails.typeNodeId = UA_NODEID_NUMERIC(0, UA_NS0ID_BASEDATAVARIABLETYPE);
                break;

            case UA_VALUERANK_ONE_DIMENSION:
                vtAttr = UA_VariableTypeAttributes_default;
                UA_NodeId_copy(&UA_TYPES[vDetails.type].typeId, &vtAttr.dataType);
                vtAttr.valueRank = vDetails.valueRank;
                vtAttr.arrayDimensions = vDetails.arrayDimensions;
                vtAttr.arrayDimensionsSize = vDetails.arrayDimensionsSize;
                vtAttr.displayName = UA_LOCALIZEDTEXT("en-US", vDetails.typeName);
                UA_Variant_setArray(&vtAttr.value, vDetails.value,
                                    vDetails.arrayDimensions[0], &UA_TYPES[vDetails.type]);
                UA_Server_addVariableTypeNode(server, UA_NODEID_NULL,
                                              UA_NODEID_NUMERIC(0, UA_NS0ID_BASEDATAVARIABLETYPE),
                                              UA_NODEID_NUMERIC(0, UA_NS0ID_HASSUBTYPE),
                                              UA_QUALIFIEDNAME(1, vDetails.typeName),
                                              UA_NODEID_NULL, vtAttr, NULL, &vDetails.typeNodeId);
                break;

            default:
                UA_LOG_ERROR(pubsubLogger, UA_LOGCATEGORY_USERLAND, "Valuerank not handled");
                break;
        }
    }
    return;
}

static UA_Server*
setServer(UA_String *transportProfile,
          UA_NetworkAddressUrlDataType *networkAddressUrl, UA_UInt32 id) {
    UA_Server *server;

    server = UA_Server_new();
    UA_ServerConfig *config = UA_Server_getConfig(server);
    UA_ServerConfig_setDefault(config);
    UA_ServerConfig_addPubSubTransportLayer(config, UA_PubSubTransportLayerUDPMP());

    addPubSubConnection(server, transportProfile, networkAddressUrl, id);
    return server;
}

static UA_StatusCode
subscribe(UA_Server *server, InstanceData *instanceArray, UA_UInt32 id,
          UA_UInt32 nbReader, UA_UInt32 maxVariableNb, UA_Duration interval,
          UA_UInt16 (*get_reader_id)(UA_UInt32 nb),
          void (*update)(UA_UInt32 id, const UA_DataValue*, bool print)) {
    UA_UInt16 publisherIdent;
    UA_StatusCode retval;
    char readerName[19];

    callbackUpdate = update;

    retval = addReaderGroup(server);
    if (retval != UA_STATUSCODE_GOOD)
        return EXIT_FAILURE;

    memset (&readerConfig, 0, sizeof(UA_DataSetReaderConfig));
    readerConfig.publisherId.type = &UA_TYPES[UA_TYPES_UINT16];
    readerConfig.writerGroupId    = WRITER_GROUP_ID;
    readerConfig.dataSetWriterId  = DATA_SET_WRITER_ID;

    for (UA_UInt32 i = 0; i < nbReader; i++) {
        publisherIdent = get_reader_id(i);
        UA_snprintf(readerName, sizeof(readerName) - 1, "DataSet Reader %d", publisherIdent);
        readerConfig.name = UA_STRING(readerName);
        readerConfig.publisherId.data = &publisherIdent;

        retval = addDataSetReader(server, instanceArray[i].variables,
                                  publisherIdent);
        if (retval != UA_STATUSCODE_GOOD)
            return EXIT_FAILURE;

        /* Add SubscribedVariables to the created DataSetReader */
        retval = addSubscribedVariables(server, readerIdent,
                                        instanceArray[i].variables.variableArray,
                                        i, maxVariableNb, interval,
                                        instanceArray[i].init_node_id);
        if (retval != UA_STATUSCODE_GOOD)
            return EXIT_FAILURE;
    }

    return retval;
}

int runPubsub(const UA_Logger *logger, UA_String *transportProfile,
              UA_NetworkAddressUrlDataType *networkAddressUrl,
              VariableStruct variables, UA_UInt32 id,
              InstanceData *readerArray, UA_UInt32 nbReader,
              UA_UInt32 maxVariableNb, UA_Duration interval,
              UA_UInt16 (*get_reader_id)(UA_UInt32 nb),
              VariableData (*get_value)(UA_String identifier),
              void (*update)(UA_UInt32 id, const UA_DataValue*, bool print),
              UA_Boolean *running) {
    UA_Server *server;
    UA_StatusCode retval;
    pubsubLogger = logger;

    server = setServer(transportProfile, networkAddressUrl, id);
    setVariableType(server, variables);

    /* Publishing */

    pubsubGetValue = get_value;

    addPublishedDataSet(server, id);
    for(UA_UInt32 i = 0; i < variables.nbVariable; i++) {
        retval = addDataSourceVariable(server, variables.variableArray[i]);
        if (retval != UA_STATUSCODE_GOOD)
            return EXIT_FAILURE;
        addDataSetField(server, variables.variableArray[i]);
    }

    addWriterGroup(server, interval);
    retval = addDataSetWriter(server);
    if (retval != UA_STATUSCODE_GOOD)
        return EXIT_FAILURE;

    /* Subscribing */

    subscribe(server, readerArray, id, nbReader, maxVariableNb, interval,
              get_reader_id, update);

    retval = UA_Server_run(server, running);
    UA_Server_delete(server);
    return retval == UA_STATUSCODE_GOOD ? EXIT_SUCCESS : EXIT_FAILURE;
}
